@@ -1,14 +1,13 @@
-//! [![github]](https://github.com/dtolnay/trybuild)&ensp;[![crates-io]](https://crates.io/crates/trybuild)&ensp;[![docs-rs]](https://docs.rs/trybuild)
-//!
-//! [github]: https://img.shields.io/badge/github-8da0cb?style=for-the-badge&labelColor=555555&logo=github
-//! [crates-io]: https://img.shields.io/badge/crates.io-fc8d62?style=for-the-badge&labelColor=555555&logo=rust
-//! [docs-rs]: https://img.shields.io/badge/docs.rs-66c2a5?style=for-the-badge&labelColor=555555&logo=docs.rs
+//! [![github]](https://github.com/GuillaumeGomez/trybuild2)&ensp;[![crates-io]](https://crates.io/crates/trybuild2)&ensp;[![docs-rs]](https://docs.rs/trybuild2)
 //!
 //! <br>
 //!
+//! `trybuild2` is a fork of [trybuild](https://github.com/dtolnay/trybuild)
+//! which allows to have inline tests.
+//!
 //! #### &emsp;A compiler diagnostics testing library in just 3 functions.
 //!
-//! Trybuild is a test harness for invoking rustc on a set of test cases and
+//! Trybuild2 is a test harness for invoking rustc on a set of test cases and
 //! asserting that any resulting error messages are the ones intended.
 //!
 //! Such tests are commonly useful for testing error reporting involving
@@ -21,19 +20,19 @@
 //! aspects of the user's interaction with a library outside of what would be
 //! covered by ordinary API tests.
 //!
-//! Nothing here is specific to macros; trybuild would work equally well for
+//! Nothing here is specific to macros; trybuild2 would work equally well for
 //! testing misuse of non-macro APIs.
 //!
 //! <br>
 //!
 //! # Compile-fail tests
 //!
-//! A minimal trybuild setup looks like this:
+//! A minimal trybuild2 setup looks like this:
 //!
 //! ```
 //! #[test]
 //! fn ui() {
-//!     let t = trybuild::TestCases::new();
+//!     let t = trybuild2::TestCases::new();
 //!     t.compile_fail("tests/ui/*.rs");
 //! }
 //! ```
@@ -80,7 +79,7 @@
 //! ```
 //! #[test]
 //! fn ui() {
-//!     let t = trybuild::TestCases::new();
+//!     let t = trybuild2::TestCases::new();
 //!     t.pass("tests/01-parse-header.rs");
 //!     t.pass("tests/02-parse-body.rs");
 //!     t.compile_fail("tests/03-expand-four-errors.rs");
@@ -89,6 +88,8 @@
 //!     //t.pass("tests/06-make-work-in-function.rs");
 //!     //t.pass("tests/07-init-array.rs");
 //!     //t.compile_fail("tests/08-ident-span.rs");
+//!     t.compile_fail_inline("name", "fn main() {}", "path-to-stderr");
+//!     t.compile_fail_check_sub("tests/03-expand-four-errors.rs", "I want to find this!");
 //! }
 //! ```
 //!
@@ -124,7 +125,7 @@
 //! </p>
 //!
 //! Alternatively, run `cargo test` with the environment variable
-//! `TRYBUILD=overwrite` to skip the *wip* directory and write all compiler
+//! `TRYBUILD2=overwrite` to skip the *wip* directory and write all compiler
 //! output directly in place. You'll want to check `git diff` afterward to be
 //! sure the compiler's output is what you had in mind.
 //!
@@ -208,55 +209,35 @@
 //! makes error messages that we care about substantially worse, it is also
 //! important to catch and report as a compiler issue.
 //!
-//! <br>
+//! # Inline checks
 //!
-//! # Troubleshooting
+//! It's possible to provide a code to compile directly through
+//! `compile_fail_inline` and `pass_inline`. It allows you to have some more
+//! generic checks which can be generated at runtime if needed.
 //!
-//! The Rust compiler's diagnostic output can vary as a function of whether the
-//! `rust-src` Rustup component is installed. The compiler will render source
-//! snippets from the standard library if the standard library source is
-//! available locally, and will simply omit snippets if not. This can account
-//! for differences between CI and local development.
+//! # Sub-string checks
 //!
-//! If you have compile_fail tests pertaining to standard library traits or
-//! types, you can ensure a consistent environment by adding a
-//! rust-toolchain.toml file with the following content.
-//!
-//! ```toml
-//! [toolchain]
-//! components = ["rust-src"]
-//! ```
+//! If you need  more control over the stderr check as well, you can take a
+//! look at `compile_fail_check_sub` and `compile_fail_inline_check_sub`.
 
-#![doc(html_root_url = "https://docs.rs/trybuild/1.0.105")]
-#![cfg_attr(not(check_cfg), allow(unexpected_cfgs))]
 #![allow(
     clippy::collapsible_if,
-    clippy::comparison_chain,
     clippy::default_trait_access,
-    clippy::derive_partial_eq_without_eq,
     clippy::doc_markdown,
-    clippy::elidable_lifetime_names,
     clippy::enum_glob_use,
     clippy::iter_not_returning_iterator, // https://github.com/rust-lang/rust-clippy/issues/8285
-    clippy::let_underscore_untyped, // https://github.com/rust-lang/rust-clippy/issues/10410
+    clippy::let_underscore_drop,
     clippy::manual_assert,
     clippy::manual_range_contains,
-    clippy::module_inception,
     clippy::module_name_repetitions,
     clippy::must_use_candidate,
-    clippy::needless_lifetimes,
     clippy::needless_pass_by_value,
     clippy::non_ascii_literal,
     clippy::range_plus_one,
-    clippy::similar_names,
     clippy::single_match_else,
-    clippy::test_attr_in_doctest,
     clippy::too_many_lines,
     clippy::trivially_copy_pass_by_ref,
-    clippy::uninhabited_references,
-    clippy::uninlined_format_args,
-    clippy::unused_self,
-    clippy::while_let_on_iterator,
+    clippy::unused_self
 )]
 #![deny(clippy::clone_on_ref_ptr)]
 
@@ -272,10 +253,8 @@ mod diff;
 mod directory;
 mod env;
 mod error;
-mod expand;
 mod features;
 mod flock;
-mod inherit;
 mod manifest;
 mod message;
 mod normalize;
@@ -299,14 +278,60 @@ struct Runner {
 
 #[derive(Clone, Debug)]
 struct Test {
-    path: PathBuf,
     expected: Expected,
+    inner: TestKind,
 }
 
-#[derive(Copy, Clone, Debug)]
+#[derive(Clone, Debug)]
+enum TestKind {
+    File(FileTest),
+    Inline(InlineTest),
+}
+
+impl Test {
+    fn is_inline(&self) -> bool {
+        // It's because `matches` isn't available in 1.36... (MSRV)
+        match self.inner {
+            TestKind::Inline(_) => true,
+            TestKind::File(_) => false,
+        }
+    }
+
+    fn stderr_path(&self) -> PathBuf {
+        match self.inner {
+            TestKind::Inline(InlineTest {
+                ref stderr_path, ..
+            }) => stderr_path.clone().expect("not a compile fail test"),
+            TestKind::File(FileTest { ref path, .. }) => path.with_extension("stderr"),
+        }
+    }
+
+    fn path(&self) -> &PathBuf {
+        match self.inner {
+            TestKind::File(FileTest { ref path, .. })
+            | TestKind::Inline(InlineTest { ref path, .. }) => path,
+        }
+    }
+}
+
+#[derive(Clone, Debug)]
+struct FileTest {
+    path: PathBuf,
+}
+
+#[derive(Clone, Debug)]
+struct InlineTest {
+    code: String,
+    name: String,
+    path: PathBuf,
+    stderr_path: Option<PathBuf>,
+}
+
+#[derive(Clone, Debug)]
 enum Expected {
     Pass,
     CompileFail,
+    CompileFailSubString(String),
 }
 
 impl TestCases {
@@ -319,15 +344,64 @@ impl TestCases {
 
     pub fn pass<P: AsRef<Path>>(&self, path: P) {
         self.runner.borrow_mut().tests.push(Test {
-            path: path.as_ref().to_owned(),
             expected: Expected::Pass,
+            inner: TestKind::File(FileTest {
+                path: path.as_ref().to_owned(),
+            }),
+        });
+    }
+
+    pub fn pass_inline(&self, name: &str, code: &str) {
+        self.runner.borrow_mut().tests.push(Test {
+            expected: Expected::Pass,
+            inner: TestKind::Inline(InlineTest {
+                code: code.to_owned(),
+                name: name.to_owned(),
+                path: PathBuf::from(name),
+                stderr_path: None,
+            }),
         });
     }
 
     pub fn compile_fail<P: AsRef<Path>>(&self, path: P) {
         self.runner.borrow_mut().tests.push(Test {
-            path: path.as_ref().to_owned(),
             expected: Expected::CompileFail,
+            inner: TestKind::File(FileTest {
+                path: path.as_ref().to_owned(),
+            }),
+        });
+    }
+
+    pub fn compile_fail_inline<P: AsRef<Path>>(&self, name: &str, code: &str, stderr_path: P) {
+        self.runner.borrow_mut().tests.push(Test {
+            expected: Expected::CompileFail,
+            inner: TestKind::Inline(InlineTest {
+                code: code.to_owned(),
+                name: name.to_owned(),
+                path: PathBuf::from(name),
+                stderr_path: Some(stderr_path.as_ref().to_owned()),
+            }),
+        });
+    }
+
+    pub fn compile_fail_check_sub<P: AsRef<Path>>(&self, path: P, sub_string: &str) {
+        self.runner.borrow_mut().tests.push(Test {
+            expected: Expected::CompileFailSubString(sub_string.to_owned()),
+            inner: TestKind::File(FileTest {
+                path: path.as_ref().to_owned(),
+            }),
+        });
+    }
+
+    pub fn compile_fail_inline_check_sub(&self, name: &str, code: &str, sub_string: &str) {
+        self.runner.borrow_mut().tests.push(Test {
+            expected: Expected::CompileFailSubString(sub_string.to_owned()),
+            inner: TestKind::Inline(InlineTest {
+                code: code.to_owned(),
+                name: name.to_owned(),
+                path: PathBuf::from(name),
+                stderr_path: None,
+            }),
         });
     }
 }
