@@ -7,15 +7,14 @@ use crate::flock::Lock;
 use crate::manifest::{Bin, Build, Config, Manifest, Name, Package, Workspace};
 use crate::message::{self, Fail, Warn};
 use crate::normalize::{self, Context, Variations};
-use crate::{features, rustflags, Expected, Runner, Test};
-use serde_derive::Deserialize;
+use crate::{features, rustflags, Expected, FileTest, InlineTest, Runner, Test, TestKind};
 use std::collections::BTreeMap as Map;
 use std::env;
 use std::ffi::{OsStr, OsString};
-use std::fs::{self, File};
+use std::fs::{self, File, OpenOptions};
+use std::io::Write;
 use std::mem;
 use std::path::{Path, PathBuf};
-use std::str;
 
 #[derive(Debug)]
 pub struct Project {
@@ -30,7 +29,6 @@ pub struct Project {
     pub workspace: Directory,
     pub path_dependencies: Vec<PathDependency>,
     manifest: Manifest,
-    pub keep_going: bool,
 }
 
 #[derive(Debug)]
@@ -45,9 +43,9 @@ impl Runner {
         filter(&mut tests);
 
         let (project, _lock) = (|| {
-            let mut project = self.prepare(&tests)?;
+            let project = self.prepare(&tests)?;
             let lock = Lock::acquire(path!(project.dir / ".lock"));
-            self.write(&mut project)?;
+            self.write(&project)?;
             Ok((project, lock))
         })()
         .unwrap_or_else(|err| {
@@ -62,14 +60,6 @@ impl Runner {
 
         if tests.is_empty() {
             message::no_tests_enabled();
-        } else if project.keep_going && !project.has_pass {
-            failures = match self.run_all(&project, tests) {
-                Ok(failures) => failures,
-                Err(err) => {
-                    message::test_fail(err);
-                    len
-                }
-            }
         } else {
             for test in tests {
                 if let Err(err) = test.run(&project) {
@@ -81,7 +71,7 @@ impl Runner {
 
         print!("\n\n");
 
-        if failures > 0 && project.name != "trybuild-tests" {
+        if failures > 0 && project.name != "trybuild2-tests" {
             panic!("{} of {} tests failed", failures, len);
         }
     }
@@ -100,7 +90,9 @@ impl Runner {
         for e in tests {
             match e.test.expected {
                 Expected::Pass => has_pass = true,
-                Expected::CompileFail => has_compile_fail = true,
+                Expected::CompileFailSubString(_) | Expected::CompileFail => {
+                    has_compile_fail = true;
+                }
             }
         }
 
@@ -136,6 +128,7 @@ impl Runner {
             crate_name,
             &workspace,
             &project_name,
+            &project_dir,
             &source_dir,
             tests,
             source_manifest,
@@ -157,11 +150,10 @@ impl Runner {
             workspace,
             path_dependencies,
             manifest,
-            keep_going: false,
         })
     }
 
-    fn write(&self, project: &mut Project) -> Result<()> {
+    fn write(&self, project: &Project) -> Result<()> {
         let manifest_toml = toml::to_string(&project.manifest)?;
 
         let config = self.make_config();
@@ -170,23 +162,20 @@ impl Runner {
         fs::create_dir_all(path!(project.dir / ".cargo"))?;
         fs::write(path!(project.dir / ".cargo" / "config"), config_toml)?;
         fs::write(path!(project.dir / "Cargo.toml"), manifest_toml)?;
-
-        let main_rs = b"\
-            #![allow(unknown_lints, unused_crate_dependencies, missing_docs)]\n\
-            fn main() {}\n\
-        ";
-        fs::write(path!(project.dir / "main.rs"), &main_rs[..])?;
+        fs::write(path!(project.dir / "main.rs"), b"fn main() {}\n")?;
 
         cargo::build_dependencies(project)?;
 
         Ok(())
     }
 
+    #[allow(clippy::too_many_arguments)]
     fn make_manifest(
         &self,
         crate_name: String,
         workspace: &Directory,
         project_name: &str,
+        project_dir: &Directory,
         source_dir: &Directory,
         tests: &[ExpandedTest],
         source_manifest: dependencies::Manifest,
@@ -251,9 +240,13 @@ impl Runner {
 
         for expanded in tests {
             if expanded.error.is_none() {
+                let path = match expanded.test.inner {
+                    TestKind::File(ref t) => source_dir.join(&t.path),
+                    TestKind::Inline(ref t) => project_dir.join(&format!("{}.rs", t.name)),
+                };
                 manifest.bins.push(Bin {
                     name: expanded.name.clone(),
-                    path: source_dir.join(&expanded.test.path),
+                    path: source_dir.join(&path),
                 });
             }
         }
@@ -268,77 +261,41 @@ impl Runner {
             },
         }
     }
-
-    fn run_all(&self, project: &Project, tests: Vec<ExpandedTest>) -> Result<usize> {
-        let mut failures = 0;
-        let output = cargo::build_all_tests(project)?;
-        let parsed = parse_cargo_json(&output.stdout);
-        let fallback = Stderr::default();
-
-        for mut t in tests {
-            let show_expected = false;
-            message::begin_test(&t.test, show_expected);
-
-            if t.error.is_none() {
-                t.error = check_exists(&t.test.path).err();
-            }
-
-            if t.error.is_none() {
-                let src_path = project.source_dir.join(&t.test.path);
-                let this_test = parsed.stderrs.get(&src_path).unwrap_or(&fallback);
-                t.error = t.test.check(project, &t.name, this_test, "").err();
-            }
-
-            if let Some(err) = t.error {
-                failures += 1;
-                message::test_fail(err);
-            }
-        }
-
-        Ok(failures)
-    }
 }
 
 impl Test {
     fn run(&self, project: &Project, name: &Name) -> Result<()> {
         let show_expected = project.has_pass && project.has_compile_fail;
         message::begin_test(self, show_expected);
-        check_exists(&self.path)?;
+        match self.inner {
+            TestKind::File(FileTest { ref path }) => check_exists(path)?,
+            TestKind::Inline(ref t) => create_inline_test(t, project)?,
+        };
 
         let output = cargo::build_test(project, name)?;
-        let parsed = parse_cargo_json(&output.stdout);
-        let src_path = project.source_dir.join(&self.path);
-        let fallback = Stderr::default();
-        let this_test = parsed.stderrs.get(&src_path).unwrap_or(&fallback);
-        self.check(project, name, this_test, &parsed.stdout)
-    }
-
-    fn check(
-        &self,
-        project: &Project,
-        name: &Name,
-        result: &Stderr,
-        build_stdout: &str,
-    ) -> Result<()> {
-        let success = result.success;
+        let success = output.status.success();
+        let stdout = output.stdout;
         let stderr = normalize::diagnostics(
-            &result.stderr,
+            output.stderr,
             Context {
                 krate: &name.0,
                 source_dir: &project.source_dir,
                 workspace: &project.workspace,
-                input_file: &self.path,
+                input_file: self.path(),
                 target_dir: &project.target_dir,
                 path_dependencies: &project.path_dependencies,
             },
         );
 
-        let check = match self.expected {
-            Expected::Pass => Test::check_pass,
-            Expected::CompileFail => Test::check_compile_fail,
-        };
-
-        check(self, project, name, success, build_stdout, stderr)
+        match self.expected {
+            Expected::Pass => Test::check_pass(self, project, name, success, stdout, stderr),
+            Expected::CompileFail => {
+                Test::check_compile_fail(self, project, name, success, stdout, stderr)
+            }
+            Expected::CompileFailSubString(ref s) => {
+                Test::check_compile_fail_sub_str(self, name, success, stdout, stderr, s)
+            }
+        }
     }
 
     fn check_pass(
@@ -346,7 +303,7 @@ impl Test {
         project: &Project,
         name: &Name,
         success: bool,
-        build_stdout: &str,
+        build_stdout: Vec<u8>,
         variations: Variations,
     ) -> Result<()> {
         let preferred = variations.preferred();
@@ -356,7 +313,7 @@ impl Test {
         }
 
         let mut output = cargo::run_test(project, name)?;
-        output.stdout.splice(..0, build_stdout.bytes());
+        output.stdout.splice(..0, build_stdout);
         message::output(preferred, &output);
         if output.status.success() {
             Ok(())
@@ -370,19 +327,12 @@ impl Test {
         project: &Project,
         _name: &Name,
         success: bool,
-        build_stdout: &str,
+        build_stdout: Vec<u8>,
         variations: Variations,
     ) -> Result<()> {
-        let preferred = variations.preferred();
+        let preferred = check_success(success, &build_stdout, &variations)?;
 
-        if success {
-            message::should_not_have_compiled();
-            message::fail_output(Fail, build_stdout);
-            message::warnings(preferred);
-            return Err(Error::ShouldNotHaveCompiled);
-        }
-
-        let stderr_path = self.path.with_extension("stderr");
+        let stderr_path = self.stderr_path();
 
         if !stderr_path.exists() {
             match project.update {
@@ -395,15 +345,15 @@ impl Test {
                         .file_name()
                         .unwrap_or_else(|| OsStr::new("test.stderr"));
                     let wip_path = wip_dir.join(stderr_name);
-                    message::write_stderr_wip(&wip_path, &stderr_path, preferred);
-                    fs::write(wip_path, preferred).map_err(Error::WriteStderr)?;
+                    message::write_stderr_wip(&wip_path, &stderr_path, &preferred);
+                    fs::write(wip_path, &preferred).map_err(Error::WriteStderr)?;
                 }
                 Update::Overwrite => {
-                    message::overwrite_stderr(&stderr_path, preferred);
-                    fs::write(stderr_path, preferred).map_err(Error::WriteStderr)?;
+                    message::overwrite_stderr(&stderr_path, &preferred);
+                    fs::write(stderr_path, &preferred).map_err(Error::WriteStderr)?;
                 }
             }
-            message::fail_output(Warn, build_stdout);
+            message::fail_output(Warn, &build_stdout);
             return Ok(());
         }
 
@@ -418,16 +368,50 @@ impl Test {
 
         match project.update {
             Update::Wip => {
-                message::mismatch(&expected, preferred);
+                message::mismatch(&expected, &preferred, "");
                 Err(Error::Mismatch)
             }
             Update::Overwrite => {
-                message::overwrite_stderr(&stderr_path, preferred);
-                fs::write(stderr_path, preferred).map_err(Error::WriteStderr)?;
+                message::overwrite_stderr(&stderr_path, &preferred);
+                fs::write(stderr_path, &preferred).map_err(Error::WriteStderr)?;
                 Ok(())
             }
         }
     }
+
+    fn check_compile_fail_sub_str(
+        &self,
+        _name: &Name,
+        success: bool,
+        build_stdout: Vec<u8>,
+        variations: Variations,
+        expected: &str,
+    ) -> Result<()> {
+        let preferred = check_success(success, &build_stdout, &variations)?;
+
+        let expected = expected.replace("\r\n", "\n");
+
+        if variations.any(|stderr| stderr.contains(&expected)) {
+            message::ok();
+            Ok(())
+        } else {
+            message::mismatch(&format!("{}\n", expected), &preferred, " SUBSTRING TO FIND");
+            Err(Error::Mismatch)
+        }
+    }
+}
+
+fn check_success(success: bool, build_stdout: &[u8], variations: &Variations) -> Result<String> {
+    let preferred = variations.preferred();
+
+    if success {
+        message::should_not_have_compiled();
+        message::fail_output(Fail, build_stdout);
+        message::warnings(preferred);
+        Err(Error::ShouldNotHaveCompiled)
+    } else {
+        Ok(preferred.to_owned())
+    }
 }
 
 fn check_exists(path: &Path) -> Result<()> {
@@ -440,6 +424,17 @@ fn check_exists(path: &Path) -> Result<()> {
     }
 }
 
+fn create_inline_test(test: &InlineTest, project: &Project) -> Result<()> {
+    let path = path!(project.dir / format!("{}.rs", test.name));
+    let mut file = OpenOptions::new()
+        .write(true)
+        .truncate(true)
+        .create(true)
+        .open(&path)
+        .map_err(Error::FileCreation)?;
+    write!(file, "{}", test.code).map_err(Error::FileCreation)
+}
+
 #[derive(Debug)]
 struct ExpandedTest {
     name: Name,
@@ -457,37 +452,46 @@ fn expand_globs(tests: &[Test]) -> Vec<ExpandedTest> {
     }
 
     fn bin_name(i: usize) -> Name {
-        Name(format!("trybuild{:03}", i))
+        Name(format!("trybuild2{:03}", i))
     }
 
     let mut vec = Vec::new();
 
     for test in tests {
-        let mut expanded = ExpandedTest {
-            name: bin_name(vec.len()),
-            test: test.clone(),
-            error: None,
-        };
-        if let Some(utf8) = test.path.to_str() {
-            if utf8.contains('*') {
-                match glob(utf8) {
-                    Ok(paths) => {
-                        for path in paths {
-                            vec.push(ExpandedTest {
-                                name: bin_name(vec.len()),
-                                test: Test {
-                                    path,
-                                    expected: expanded.test.expected,
-                                },
-                                error: None,
-                            });
+        let expanded = if test.is_inline() {
+            ExpandedTest {
+                name: bin_name(vec.len()),
+                test: test.clone(),
+                error: None,
+            }
+        } else {
+            let mut expanded = ExpandedTest {
+                name: bin_name(vec.len()),
+                test: test.clone(),
+                error: None,
+            };
+            if let Some(utf8) = test.path().to_str() {
+                if utf8.contains('*') {
+                    match glob(utf8) {
+                        Ok(paths) => {
+                            for path in paths {
+                                vec.push(ExpandedTest {
+                                    name: bin_name(vec.len()),
+                                    test: Test {
+                                        expected: expanded.test.expected.clone(),
+                                        inner: TestKind::File(FileTest { path }),
+                                    },
+                                    error: None,
+                                });
+                            }
+                            continue;
                         }
-                        continue;
+                        Err(error) => expanded.error = Some(error),
                     }
-                    Err(error) => expanded.error = Some(error),
                 }
             }
-        }
+            expanded
+        };
         vec.push(expanded);
     }
 
@@ -507,13 +511,13 @@ impl ExpandedTest {
     }
 }
 
-// Filter which test cases are run by trybuild.
+// Filter which test cases are run by trybuild2.
 //
-//     $ cargo test -- ui trybuild=tuple_structs.rs
+//     $ cargo test -- ui trybuild2=tuple_structs.rs
 //
-// The first argument after `--` must be the trybuild test name i.e. the name of
-// the function that has the #[test] attribute and calls trybuild. That's to get
-// Cargo to run the test at all. The next argument starting with `trybuild=`
+// The first argument after `--` must be the trybuild2 test name i.e. the name of
+// the function that has the #[test] attribute and calls trybuild2. That's to get
+// Cargo to run the test at all. The next argument starting with `trybuild2=`
 // provides a filename filter. Only test cases whose filename contains the
 // filter string will be run.
 #[allow(clippy::needless_collect)] // false positive https://github.com/rust-lang/rust-clippy/issues/5991
@@ -521,7 +525,7 @@ fn filter(tests: &mut Vec<ExpandedTest>) {
     let filters = env::args_os()
         .flat_map(OsString::into_string)
         .filter_map(|mut arg| {
-            const PREFIX: &str = "trybuild=";
+            const PREFIX: &str = "trybuild2=";
             if arg.starts_with(PREFIX) && arg != PREFIX {
                 Some(arg.split_off(PREFIX.len()))
             } else {
@@ -537,86 +541,6 @@ fn filter(tests: &mut Vec<ExpandedTest>) {
     tests.retain(|t| {
         filters
             .iter()
-            .any(|f| t.test.path.to_string_lossy().contains(f))
+            .any(|f| t.test.path().to_string_lossy().contains(f))
     });
 }
-
-#[derive(Deserialize)]
-struct CargoMessage {
-    #[allow(dead_code)]
-    reason: Reason,
-    target: RustcTarget,
-    message: RustcMessage,
-}
-
-#[derive(Deserialize)]
-enum Reason {
-    #[serde(rename = "compiler-message")]
-    CompilerMessage,
-}
-
-#[derive(Deserialize)]
-struct RustcTarget {
-    src_path: PathBuf,
-}
-
-#[derive(Deserialize)]
-struct RustcMessage {
-    rendered: String,
-    level: String,
-}
-
-struct ParsedOutputs {
-    stdout: String,
-    stderrs: Map<PathBuf, Stderr>,
-}
-
-struct Stderr {
-    success: bool,
-    stderr: String,
-}
-
-impl Default for Stderr {
-    fn default() -> Self {
-        Stderr {
-            success: true,
-            stderr: String::new(),
-        }
-    }
-}
-
-fn parse_cargo_json(stdout: &[u8]) -> ParsedOutputs {
-    let mut map = Map::new();
-    let mut nonmessage_stdout = String::new();
-    let mut remaining = &*String::from_utf8_lossy(stdout);
-    while !remaining.is_empty() {
-        let begin = match remaining.find("{\"reason\":") {
-            Some(begin) => begin,
-            None => break,
-        };
-        let (nonmessage, rest) = remaining.split_at(begin);
-        nonmessage_stdout.push_str(nonmessage);
-        let len = match rest.find('\n') {
-            Some(end) => end + 1,
-            None => rest.len(),
-        };
-        let (message, rest) = rest.split_at(len);
-        if let Ok(de) = serde_json::from_str::<CargoMessage>(message) {
-            if de.message.level != "failure-note" {
-                let mut entry = map
-                    .entry(de.target.src_path)
-                    .or_insert_with(Stderr::default);
-                if de.message.level == "error" {
-                    entry.success = false;
-                }
-                entry.stderr.push_str(&de.message.rendered);
-            }
-        }
-        remaining = rest;
-    }
-    nonmessage_stdout.push_str(remaining);
-    ParsedOutputs {
-        stdout: nonmessage_stdout,
-        stderrs: map,
-    }
-}
